==> newsstar-1.5.3.tar.gz: newsstar-1.5.3/src/intercom.c <==
/*
	newsstar - advanced news fetcher
	Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <limits.h>

#include "deblog.h"
#include "intercom.h"

#if USE_PTHREADS

static char *rstrip(const char *s)
{
	char *s2 = XSTRDUP(s);
	size_t l = strlen(s2);
	
	if (s2[l - 1] == '\n')
		s2[l - 1] = 0;
	return s2;
}

IntercomQ *intercom_queue_new(int fd, BOOL threadsafe)
{
	IntercomQ *q;
	
	UNUSED(fd);
	UNUSED(threadsafe);
	
	q = XMALLOC(sizeof(IntercomQ));
	q->queue = NULL;
	q->qsize = q->qlen = 0;
	q->id = NULL;
	pthread_cond_init(&q->cond, NULL);
	pthread_mutex_init(&q->mutex, NULL);
	return q;
}

void intercom_queue_delete(IntercomQ *q)
{
	if (!q)
		return;
	pthread_cond_destroy(&q->cond);
	pthread_mutex_destroy(&q->mutex);
	free(q->queue);
	free(q);
}

char *intercom_queue_get_command(IntercomQ *q)
{
	char *result;
	char *stripped;
	
	pthread_mutex_lock(&q->mutex);
	if (!q->qlen)
	{
		deblog_debug("IQ: Waiting for cond on q %p\n", q);
		pthread_cond_wait(&q->cond, &q->mutex);
	}
	result = q->queue[0];
	stripped = rstrip(result);
	if (--q->qlen)
	{
		int n;
		
		for (n = 0; n < q->qlen; ++n)
			q->queue[n] = q->queue[n + 1];
	}
	deblog_debug("IQ: Read command %s from q %p, qlen %ld\n",
			stripped, q, q->qlen);
	free(stripped);
	pthread_mutex_unlock(&q->mutex);
	return result;
}

#else /* !USE_PTHREADS */

IntercomQ *intercom_queue_new(int fd, BOOL threadsafe)
{
	return rw_create_buf(fd, FALSE, threadsafe, -1);
}

void intercom_queue_delete(IntercomQ *q)
{
	if (!q)
		return;
	rw_free_buf(q);
}

#endif /* !USE_PTHREADS */


Intercom *intercom_new(IntercomQ *q)
{
	Intercom *ic = XMALLOC(sizeof(Intercom));

	ic->queue = q;
	ic->data.data = NULL;
#if USE_PTHREADS
	q->id = &ic->data;
#endif
	return ic;
}

void intercom_delete(Intercom *ic)
{
	if (!ic)
		return;
	free(ic->data.data);
	free(ic);
}

void intercom_send_command(IntercomQ *q, intercom_Command code, int thread_idx,
		const char *data, ...)
{
	va_list ap;
	
	va_start(ap, data);
	intercom_send_vcommand(q, code, thread_idx, data, ap);
	va_end(ap);
}

#define INTERCOM_ATOMIC PIPE_BUF
/* NB configure script checks that PIPE_BUF is large enough to be usable */

#define MAX_RECURSION 500

#if USE_PTHREADS

static void intercom_send_formatted_command(IntercomQ *q, int thread_idx,
		const char *headstr, const char *formstr, int recursion)
{
	char *longstr = XSTRDUPCAT(headstr, formstr);
	char *stripped = rstrip(longstr);
	
	pthread_mutex_lock(&q->mutex);
	if (++q->qlen > q->qsize)
	{
		q->queue = XREALLOC(q->queue, q->qlen * sizeof(char *));
		q->qsize = q->qlen;
	}
	q->queue[q->qlen - 1] = longstr;
	deblog_debug("IQ: Sending command %s on q %p, qlen %ld\n",
			stripped, q, q->qlen);
	free(stripped);
	if (q->qlen == 1)
	{
		deblog_debug("IQ: Signalling cond on q %p\n", q);
		pthread_cond_signal(&q->cond);
	}
	pthread_mutex_unlock(&q->mutex);
}

#else /* !USE_PTHREADS */

/* This makes recursion for long commands more efficient */
static void intercom_send_formatted_command(IntercomQ *q, int thread_idx,
		const char *headstr, const char *formstr, int recursion)
{
	static BOOL baled_out = FALSE;

	if (recursion > MAX_RECURSION)
	{
		fprintf(stderr,
				"\nToo much recursion in intercom_send_formatted_command, ");
		if (!baled_out)
		{
			baled_out = TRUE;
			fprintf(stderr, "exiting\n");
			fflush(stderr);
			exit(1);
		}
		else
		{
			fprintf(stderr, "aborting\n");
			fflush(stderr);
			abort();
		}
	}
	/* +2 is for CRLF terminator - should only need LF but play safe */
	if (thread_idx != -1 &&
			strlen(headstr) + strlen(formstr) + 2 >= INTERCOM_ATOMIC)
	{
		char splitheadstr[8];
		/* malloc in case it isn't a good idea to have something so large on
		 * stack */
		char *chopped = XMALLOC(INTERCOM_ATOMIC);
		size_t frag_len;

		/*
		printf("\nThread %d split:\n"
				"headstr (%d) + formstr (%d) + 2 = "
				"%d >= INTERCOM_ATOMIC (%d)\n",
				strlen(headstr), strlen(formstr),
				strlen(headstr) + strlen(formstr) + 2,
				INTERCOM_ATOMIC);
		*/
		snprintf(splitheadstr, 8, "%d %c ", thread_idx,
				command_SplitCommand);
		frag_len = INTERCOM_ATOMIC - 3 - strlen(splitheadstr);
		strncpy(chopped, formstr, frag_len);
		chopped[frag_len] = 0;
		/* Recursion, woo! */
		/*
		printf("Splitting %s %s\ninto %s %s\n%s %s\n",
				headstr, formstr,
				splitheadstr, chopped,
				headstr, formstr + frag_len);
		sleep(1);
		*/
		intercom_send_formatted_command(q, thread_idx,
				splitheadstr, chopped, recursion + 1);
		intercom_send_formatted_command(q, thread_idx,
				headstr, formstr + frag_len, recursion + 1);
	}
	else
	{
		char *longstr = XSTRDUPCAT(headstr, formstr);

		if (rw_write_line(q, longstr) <= 0)
		{
			fprintf(stderr, _("Can't send interprocess command: %s\n %s\n"),
					strerror(errno), longstr);
			/*
			perror(_("Can't send interprocess command"));
			*/
			exit(1);
		}
		free(longstr);
	}
}

#endif /* !USE_PTHREADS */

void intercom_send_vcommand(IntercomQ *q, intercom_Command code, int thread_idx,
		const char *data, va_list ap)
{
	char *formstr;
	char headstr[8];

	if (vasprintf(&formstr, data, ap) < 0)
	{
		fprintf(stderr, 
				_("Can't send interprocess command (out of memory?)\n"));
		exit(1);
	}
	snprintf(headstr, 8, "%d %c ", thread_idx, code);
	intercom_send_formatted_command(q, thread_idx, headstr, formstr, 0);
	free(formstr);
}

const intercom_data *intercom_get_command(Intercom *ic)
{
#if !USE_PTHREADS
	const
#endif
	char *line;
	size_t len;
	
	free(ic->data.data);
	ic->data.data = NULL;
#if USE_PTHREADS
	line = intercom_queue_get_command(ic->queue);
#else
	line = rw_read_line(ic->queue);
#endif
	if (!line)
	{
		deblog_error(_("Can't read interprocess command: %s\n"),
				strerror(errno));
		return NULL;
	}
	ic->data.command = command_Null;
	if (sscanf(line, "%d %c", &ic->data.thread_idx,
			(char *) &ic->data.command) < 2
			|| ic->data.command < command_First
			|| ic->data.command > command_Last)
	{
		deblog_error(_("Received bad IPC command line: %s\n"), line);
#if USE_PTHREADS
		free(line);
#endif
		free(ic->data.data);
		ic->data.data = NULL;
		return NULL;
	}
	ic->data.data = XSTRDUP(strchr(strchr(line, ' ') + 1, ' ') + 1);
#if USE_PTHREADS
	free(line);
#endif
	/* Log message may have been sent with \n replaced by \r, so switch back */
	len = strlen(ic->data.data);
	if (len > 0 && ic->data.data[len - 1] == '\r')
		ic->data.data[len - 1] = '\n';
	/*
	if (ic->data.command != command_Log)
	{
		deblog_debug(_("Received IPC line: %s\n"), line);
	}
	*/
	return &ic->data;
}

==> newsstar-1.5.6.tar.gz: newsstar-1.5.6/src/intercom.c <==
/*
	newsstar - advanced news fetcher
	Copyright (C) 2001-2002 Tony Houghton <h@realh.co.uk>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <limits.h>

#include "deblog.h"
#include "intercom.h"

#if USE_PTHREADS

static char *rstrip(const char *s)
{
	char *s2 = XSTRDUP(s);
	size_t l = strlen(s2);

	if (s2[l - 1] == '\n')
		s2[l - 1] = 0;
	return s2;
}

IntercomQ *intercom_queue_new(int fd, BOOL threadsafe)
{
	IntercomQ *q;

	UNUSED(fd);
	UNUSED(threadsafe);

	q = XMALLOC(sizeof(IntercomQ));
	q->queue = NULL;
	q->qsize = q->qlen = 0;
	q->id = NULL;
	pthread_cond_init(&q->cond, NULL);
	pthread_mutex_init(&q->mutex, NULL);
	return q;
}

void intercom_queue_delete(IntercomQ *q)
{
	if (!q)
		return;
	pthread_cond_destroy(&q->cond);
	pthread_mutex_destroy(&q->mutex);
	free(q->queue);
	free(q);
}

char *intercom_queue_get_command(IntercomQ *q)
{
	char *result;
	char *stripped;

	pthread_mutex_lock(&q->mutex);
	if (!q->qlen)
	{
		deblog_debug("IQ: Waiting for cond on q %p\n", q);
		pthread_cond_wait(&q->cond, &q->mutex);
	}
	result = q->queue[0];
	stripped = rstrip(result);
	if (--q->qlen)
	{
		int n;

		for (n = 0; n < q->qlen; ++n)
			q->queue[n] = q->queue[n + 1];
	}
	deblog_debug("IQ: Read command %s from q %p, qlen %ld\n",
			stripped, q, q->qlen);
	free(stripped);
	pthread_mutex_unlock(&q->mutex);
	return result;
}

#else /* !USE_PTHREADS */

IntercomQ *intercom_queue_new(int fd, BOOL threadsafe)
{
	return rw_create_buf(fd, FALSE, threadsafe, -1);
}

void intercom_queue_delete(IntercomQ *q)
{
	if (!q)
		return;
	rw_free_buf(q);
}

#endif /* !USE_PTHREADS */


Intercom *intercom_new(IntercomQ *q)
{
	Intercom *ic = XMALLOC(sizeof(Intercom));

	ic->queue = q;
	ic->data.data = NULL;
#if USE_PTHREADS
	q->id = &ic->data;
#endif
	return ic;
}

void intercom_delete(Intercom *ic)
{
	if (!ic)
		return;
	free(ic->data.data);
	free(ic);
}

void intercom_send_command(IntercomQ *q, intercom_Command code, int thread_idx,
		const char *data, ...)
{
	va_list ap;

	va_start(ap, data);
	intercom_send_vcommand(q, code, thread_idx, data, ap);
	va_end(ap);
}

#define INTERCOM_ATOMIC ((size_t) (PIPE_BUF - 8))
/* NB configure script checks that PIPE_BUF is large enough to be usable */

#define MAX_RECURSION 500

#if USE_PTHREADS

static void intercom_send_formatted_command(IntercomQ *q, int thread_idx,
		const char *headstr, const char *formstr, int recursion)
{
	char *longstr = XSTRDUPCAT(headstr, formstr);
	char *stripped = rstrip(longstr);

	pthread_mutex_lock(&q->mutex);
	if (++q->qlen > q->qsize)
	{
		q->queue = XREALLOC(q->queue, q->qlen * sizeof(char *));
		q->qsize = q->qlen;
	}
	q->queue[q->qlen - 1] = longstr;
	deblog_debug("IQ: Sending command %s on q %p, qlen %ld\n",
			stripped, q, q->qlen);
	free(stripped);
	if (q->qlen == 1)
	{
		deblog_debug("IQ: Signalling cond on q %p\n", q);
		pthread_cond_signal(&q->cond);
	}
	pthread_mutex_unlock(&q->mutex);
}

#else /* !USE_PTHREADS */

/* This makes recursion for long commands more efficient */
static void intercom_send_formatted_command(IntercomQ *q, int thread_idx,
		const char *headstr, const char *formstr, int recursion)
{
	static BOOL baled_out = FALSE;

	if (recursion > MAX_RECURSION)
	{
		fprintf(stderr,
				"\nToo much recursion in intercom_send_formatted_command, ");
		if (!baled_out)
		{
			baled_out = TRUE;
			fprintf(stderr, "exiting\n");
			fflush(stderr);
			exit(1);
		}
		else
		{
			fprintf(stderr, "aborting\n");
			fflush(stderr);
			abort();
		}
	}

	/* +2 is for CRLF terminator - should only need LF but play safe */
	if (thread_idx != -1 &&
			strlen(headstr) + strlen(formstr) + 2 >= INTERCOM_ATOMIC)
	{
		char splitheadstr[8];
		/* malloc in case it isn't a good idea to have something so large on
		 * stack */
		char *chopped = XMALLOC(INTERCOM_ATOMIC);
		size_t frag_len;

		/*
		printf("\nThread %d split:\n"
				"headstr (%ld) + formstr (%ld) + 2 = "
				"%ld >= INTERCOM_ATOMIC (%ld)\n",
				thread_idx,
				strlen(headstr), strlen(formstr),
				strlen(headstr) + strlen(formstr) + 2,
				INTERCOM_ATOMIC);
		*/
		snprintf(splitheadstr, 8, "%d %c ", thread_idx,
				command_SplitCommand);
		frag_len = INTERCOM_ATOMIC - 3 - strlen(splitheadstr);
		strncpy(chopped, formstr, frag_len);
		chopped[frag_len] = 0;
		/* Recursion, woo! */
		/*
		printf("Splitting %s %s\ninto %s %s\n%s %s\n",
				headstr, formstr,
				splitheadstr, chopped,
				headstr, formstr + frag_len);
		*/
		intercom_send_formatted_command(q, thread_idx,
				splitheadstr, chopped, recursion + 1);
		intercom_send_formatted_command(q, thread_idx,
				headstr, formstr + frag_len, recursion + 1);
	}
	else
	{
		char *longstr = XSTRDUPCAT(headstr, formstr);

		if (rw_write_line(q, longstr) <= 0)
		{
			fprintf(stderr, _("Can't send interprocess command: %s\n %s\n"),
					strerror(errno), longstr);
			/*
			perror(_("Can't send interprocess command"));
			*/
			exit(1);
		}
		free(longstr);
	}
}

#endif /* !USE_PTHREADS */

void intercom_send_vcommand(IntercomQ *q, intercom_Command code, int thread_idx,
		const char *data, va_list ap)
{
	char *formstr;
	char headstr[8];

	if (vasprintf(&formstr, data, ap) < 0)
	{
		fprintf(stderr,
				_("Can't send interprocess command (out of memory?)\n"));
		exit(1);
	}
	snprintf(headstr, 8, "%d %c ", thread_idx, code);
	intercom_send_formatted_command(q, thread_idx, headstr, formstr, 0);
	free(formstr);
}

const intercom_data *intercom_get_command(Intercom *ic)
{
#if !USE_PTHREADS
	const
#endif
	char *line;
	size_t len;

	free(ic->data.data);
	ic->data.data = NULL;
#if USE_PTHREADS
	line = intercom_queue_get_command(ic->queue);
#else
	line = rw_read_line(ic->queue);
#endif
	if (!line)
	{
		deblog_error(_("Can't read interprocess command: %s\n"),
				strerror(errno));
		return NULL;
	}
	char *l2 = XSTRDUP(line);
	size_t l2l = strlen(l2);
	if (l2[l2l - 1] == '\r')
	    l2[l2l - 1] = 0;
	deblog_debug("Read raw line '%s'\n", l2);
	free(l2);
	ic->data.command = command_Null;
	if (sscanf(line, "%d %c", &ic->data.thread_idx,
			(char *) &ic->data.command) < 2
			|| ic->data.command < command_First
			|| ic->data.command > command_Last)
	{
		deblog_error(_("Received bad IPC command line: %s\n"), line);
#if USE_PTHREADS
		free(line);
#endif
		free(ic->data.data);
		ic->data.data = NULL;
		return NULL;
	}
	ic->data.data = XSTRDUP(strchr(strchr(line, ' ') + 1, ' ') + 1);
#if USE_PTHREADS
	free(line);
#endif
	/* Log message may have been sent with \n replaced by \r, so switch back */
	len = strlen(ic->data.data);
	if (len > 0 && ic->data.data[len - 1] == '\r')
		ic->data.data[len - 1] = '\n';
	/*
	if (ic->data.command != command_Log)
	{
		deblog_debug(_("Received IPC line: %s\n"), line);
	}
	*/
	return &ic->data;
}
